#include <cmath>
#include <cstddef>
#include <cstdlib>
#include <iostream>
#include <limits>
#include <ostream>
#include <random>

#include <thrust/copy.h>
#include <thrust/device_vector.h>
#include <thrust/host_vector.h>

__global__ void saxpy_kernel(float* y, float* a, float* x, std::size_t N)
{
    unsigned i = blockIdx.x * blockDim.x + threadIdx.x;

    // Happens if N % blockDim.x != 0
    if (i >= N)
    {
        return;
    }

    y[i] = a[i] * x[i] + y[i];
}


int main()
{
    // Gather device specific properties first
    int device_count = 0;
    int max_threads_per_block = 0;
    cudaGetDeviceCount(&device_count);

    if (device_count > 0)
    {
        // We're lazy and assume there's only one device
        cudaDeviceProp prop;
        cudaGetDeviceProperties(&prop, 0);
        max_threads_per_block = prop.maxThreadsPerBlock;
    }
    else
    {
        std::cerr << "No devices found\n";
        return EXIT_FAILURE;
    }


    // Fill host buffers with some random arrays
    constexpr static std::size_t N = 10E6;

    thrust::host_vector<float> h_y(N);
    thrust::host_vector<float> h_a(N);
    thrust::host_vector<float> h_x(N);

    std::default_random_engine rng;
    std::uniform_real_distribution<float> dist;

    for (std::size_t i = 0; i < N; ++i)
    {
        h_y[i] = dist(rng);
        h_a[i] = dist(rng);
        h_x[i] = dist(rng);
    }
    std::cout << "Initialization... done\n";

    // Copy to the GPU
    thrust::device_vector<float> d_y(h_y);
    thrust::device_vector<float> d_a(h_a);
    thrust::device_vector<float> d_x(h_x);
    std::cout << "Copy to GPU... done\n";

    // divide and round up
    auto div_up = [](unsigned a, unsigned b)
    {   
        return (a + b - 1) / b;
    };

    // Configure grid for compute kernel
    dim3 threads_per_block = 8;

    // We may not schedule more than max_threads_per_block
    // threads per block..
    if (threads_per_block.x > max_threads_per_block)
    {
        std::cerr << "Invalid number of threads per block: "
                  << threads_per_block.x
                  << ", max: " << max_threads_per_block << '\n';
        return EXIT_FAILURE;
    }
    dim3 num_blocks = div_up(N, threads_per_block.x);

    // Measure the time it takes the kernel to execute
    // with events. (Those don't halt the whole pipeline
    // but are like tokens that are inserted into the
    // command stream.)
    cudaEvent_t start;
    cudaEvent_t stop;
    cudaEventCreate(&start);
    cudaEventCreate(&stop);

    cudaEventRecord(start);

    // Compute SAXPY on GPU
    saxpy_kernel<<<num_blocks, threads_per_block>>>(
        thrust::raw_pointer_cast(d_y.data()),
        thrust::raw_pointer_cast(d_a.data()),
        thrust::raw_pointer_cast(d_x.data()),
        N
        );

    // Synchronize on event recorded after
    // kernel execution
    cudaEventRecord(stop);
    cudaEventSynchronize(stop);
    float ms = 0.0f;
    cudaEventElapsedTime(&ms, start, stop);
    std::cout << "Elapsed: " << ms << " ms.\n";

    // Copy result back to CPU
    thrust::host_vector<float> h_y_result(d_y);

    // Compare
    for (std::size_t i = 0; i < N; ++i)
    {
        float y = h_a[i] * h_x[i] + h_y[i];

        float diff = std::abs(y - h_y_result[i]);
        if (diff > std::numeric_limits<float>::epsilon())
        {
            std::cerr << "Incorrect result\n";
            return EXIT_FAILURE;
        }
    }

    return EXIT_SUCCESS;
}
